from django.contrib import admin
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.html import format_html
from .models import Search
from .forms import SearchForm
from .filters import SearchkitFilter
from .filters import SearchableModelFilter


@admin.register(Search)
class SearchkitSearchAdmin(admin.ModelAdmin):
    form = SearchForm
    list_display = ('name', 'contenttype', 'created_date', 'apply_search_view')
    list_filter = (('contenttype', SearchableModelFilter),)

    def get_url_for_applied_search(self, obj):
        app_label = obj.contenttype.app_label
        model_name = obj.contenttype.model
        base_url = reverse(f'admin:{app_label}_{model_name}_changelist')
        return f'{base_url}?{SearchkitFilter.parameter_name}={obj.pk}'

    def response_add(self, request, obj, *args, **kwargs):
        if '_save_and_apply' in request.POST:
            return HttpResponseRedirect(self.get_url_for_applied_search(obj))
        else:
            return super().response_add(request, obj, *args, **kwargs)

    def response_change(self, request, obj, *args, **kwargs):
        if '_save_and_apply' in request.POST:
            return HttpResponseRedirect(self.get_url_for_applied_search(obj))
        else:
            return super().response_change(request, obj, *args, **kwargs)

    def apply_search_view(self, obj):
        """
        Returns a link to apply the search.
        """
        return format_html(
            '<a href="{}" >Apply search "{}"</a>',
            self.get_url_for_applied_search(obj),
            obj.name
        )
    apply_search_view.short_description = 'Apply Search'


import random
import string
import uuid
from datetime import timedelta
from decimal import Decimal
from django.utils import timezone
from django.core.management.base import BaseCommand
from django.contrib.auth.models import User
from example.models import ModelA, ModelB, ModelC, ModelD
from example.models import CHARS_CHOICES, INTEGER_CHOICES


class Command(BaseCommand):
    help = 'Generate test data for ModelA and ModelB'

    def handle(self, *args, **kwargs):
        self.stdout.write("Creating test data...")

        # Create a superuser
        if not User.objects.filter(username='admin').exists():
            User.objects.create_superuser(
                username='admin',
                email='admin@example.com',
                password='adminpassword'
            )
            self.stdout.write(self.style.SUCCESS("Superuser 'admin' created successfully!"))
        else:
            self.stdout.write(self.style.WARNING("Superuser 'admin' already exists."))

        # Create ModelC instances
        modeld_list = []
        for i in range(100):
            modeld = ModelD.objects.create(
                chars=f"ModelD chars {i}",
                integer=random.randint(1, 100),
                date=timezone.now().date() - timedelta(days=random.randint(0, 1000)),
            )
            modeld_list.append(modeld)

        # Create ModelC instances
        modelc_list = []
        for i in range(10):
            modelc = ModelC.objects.create(
                boolean=random.choice([False, True]),
                chars=f"ModelC chars {i}",
                integer=random.randint(1, 100),
                date=timezone.now().date() - timedelta(days=random.randint(0, 1000)),
            )
            modelc_list.append(modelc)

        # Create ModelB instances
        modelb_list = []
        for i in range(500):
            modelb = ModelB.objects.create(
                chars=random.choice([f"ModelB chars {i}", None]),
                integer=random.choice([random.randint(1, 1000), None]),
                decimal=random.choice([Decimal(f"{random.uniform(1, 999):.2f}"), None]),
                date=random.choice([timezone.now().date() - timedelta(days=random.randint(0, 1000)), None]),
                datetime=random.choice([timezone.now() - timedelta(days=random.randint(0, 1000)), None]),
                model_c=random.choice(modelc_list),
            )
            modelb_list.append(modelb)

        # Create ModelA instances
        modela_list = []
        for i in range(1000):
            random_string = ''.join(random.choices(string.ascii_letters, k=6))
            random_choices = random.choices(string.ascii_letters, k=1000)
            random_choices = [f'{c}\n' if i % 60 == 59 else c for i, c in enumerate(random_choices)]
            random_text = ''.join(random_choices)
            model_a = ModelA.objects.create(
                boolean=random.choice([True, False, None]),
                chars=f"ModelA chars {i}",
                chars_choices=random.choice([c[0] for c in CHARS_CHOICES]),
                text=random_text,
                email=f"user{i}-{random_string}@example.com",
                url=f"https://example.com/{random_string}/{i}",
                uuid=uuid.uuid4(),
                integer=random.randint(1, 1000),
                big_integer=random.randint(1, 100000),
                integer_choices=random.choice([c[0] for c in INTEGER_CHOICES]),
                float=random.uniform(1, 1000),
                decimal=Decimal(f"{random.uniform(1, 999):.2f}"),
                date=timezone.now().date() - timedelta(days=random.randint(0, 1000)),
                time=(timezone.now() - timedelta(minutes=random.randint(0, 1440))).time(),
                datetime=timezone.now() - timedelta(days=random.randint(0, 1000)),
            )
            modela_list.append(model_a)

        random.shuffle(modela_list)
        random.shuffle(modelb_list)
        random.shuffle(modeld_list)

        for modela in modela_list:
            modelds = random.choices(modeld_list, k=random.choice(range(5)))
            modela.model_d.add(*modelds)

        while modelb_list:
            modela = modela_list.pop()
            modela.model_b = modelb_list.pop()
            modela.save()

        self.stdout.write(self.style.SUCCESS("Test data created for ModelA, ModelB, and ModelC!"))


from django.contrib import admin
from searchkit.filters import SearchkitFilter
from .models import ModelA
from .models import ModelB
from .models import ModelC


@admin.register(ModelA)
class ModelAAdmin(admin.ModelAdmin):
    list_display = [f.name for f in ModelA._meta.fields if not f.name == 'text']
    list_filter = [SearchkitFilter]


@admin.register(ModelB)
class ModelBAdmin(admin.ModelAdmin):
    list_display = [f.name for f in ModelB._meta.fields]
    list_filter = [SearchkitFilter]


@admin.register(ModelC)
class ModelCAdmin(admin.ModelAdmin):
    list_display = [f.name for f in ModelC._meta.fields]
    list_filter = [SearchkitFilter]



from modeltree import ModelTree as BaseModelTree
from django.contrib import admin
from django.utils.translation import gettext_lazy as _


def is_searchable_model(model):
    """
    Check if the model is searchable by Searchkit.
    """
    # We do not import SearchkitFilter to avoid circular imports. So we check
    # the filter by its name.
    return (
        admin.site.is_registered(model)
        and 'SearchkitFilter' in [getattr(f, '__name__', '') for f in admin.site._registry[model].list_filter]
    )


# TODO: Make modeltree parameters configurable.
class ModelTree(BaseModelTree):
    MAX_DEPTH = 3
    FOLLOW_ACROSS_APPS = True



# Generated by Django 5.2.4 on 2025-07-15 16:37

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('example', '0002_modeld_modela_model_d'),
    ]

    operations = [
        migrations.AddField(
            model_name='modelc',
            name='boolean',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='modela',
            name='model_b',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='example.modelb'),
        ),
        migrations.AlterField(
            model_name='modelb',
            name='chars',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='modelb',
            name='date',
            field=models.DateField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='modelb',
            name='datetime',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='modelb',
            name='decimal',
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=5, null=True),
        ),
        migrations.AlterField(
            model_name='modelb',
            name='integer',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from .models import Search
from .utils import is_searchable_model


class SearchkitFilter(admin.SimpleListFilter):
    title = 'Searchkit Filter'
    parameter_name = 'search'
    template = 'searchkit/searchkit_filter.html'

    def __init__(self, request, params, model, model_admin):
        # We need the app_label and model as get parameter for the new search
        # link.
        self.searchkit_model = ContentType.objects.get_for_model(model)
        super().__init__(request, params, model, model_admin)

    def has_output(self):
        return True

    def lookups(self, request, model_admin):
        searches = Search.objects.filter(contenttype=self.searchkit_model).order_by('-created_date')
        return [(str(obj.id), obj.name) for obj in searches]

    def queryset(self, request, queryset):
        # Filter the queryset based on the selected SearchkitSearch object
        if self.value():
            search = Search.objects.get(id=int(self.value()))
            includes, excludes = search.as_lookups()
            return queryset.filter(**includes).exclude(**excludes)


class SearchableModelFilter(admin.filters.RelatedFieldListFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        contenttypes = ContentType.objects.order_by('app_label', 'model')
        self.lookup_choices = [(m.id, m) for m in contenttypes if is_searchable_model(m.model_class())]



from django import forms
from django.db import models
from django.apps import apps
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.contrib.contenttypes.models import ContentType
from django.contrib.admin import widgets
from django.contrib.admin.options import FORMFIELD_FOR_DBFIELD_DEFAULTS
from .models import Search
from .utils import ModelTree
from .utils import is_searchable_model
from . import fields as  skfields


RELOAD_CSS_CLASS = "searchkit-reload"


# TODO: Check unique_together contraint for search name and content type.
class SearchForm(forms.ModelForm):
    """
    Represents a SearchkitSearch model. Using a SearchkitFormSet for the data
    json field.
    """
    class Meta:
        model = Search
        fields = ['name']

    @property
    def media(self):
        return super().media + self.formset.media

    @cached_property
    def searchkit_model(self):
        # Try hard to get a model to work with.
        # Do we have an instance? Then use its model.
        if self.instance.pk:
            return self.instance.contenttype.model_class()

        # Otherwise check if we have a valid searchkit model form.
        elif self.searchkit_model_form.is_valid():
            return self.searchkit_model_form.cleaned_data['searchkit_model'].model_class()

        # At least check initials for a searchkit model value and use our model
        # form to validate it.
        elif 'searchkit_model' in self.searchkit_model_form.initial:
            value = self.searchkit_model_form.initial['searchkit_model']
            try:
                cleaned_value = self.searchkit_model_form.fields['searchkit_model'].clean(value)
            except forms.ValidationError:
                return None
            else:
                return cleaned_value.model_class()

    @cached_property
    def searchkit_model_form(self):
        kwargs = dict(data=self.data or None, initial=self.initial or None)
        if self.instance.pk:
            kwargs['initial'] = dict(searchkit_model=self.instance.contenttype)
        return SearchkitModelForm(**kwargs)

    @cached_property
    def formset(self):
        """
        A searchkit formset for the model.
        """
        kwargs = dict()
        if self.searchkit_model and self.data:
            kwargs = dict(data=self.data)
        elif self.searchkit_model and self.instance.pk:
            kwargs = dict(initial=self.instance.data)

        extra = 0 if self.instance.pk else 1
        formset = searchkit_formset_factory(model=self.searchkit_model, extra=extra)
        return formset(**kwargs)

    def is_valid(self):
        return self.formset.is_valid() and self.searchkit_model_form.is_valid and super().is_valid()

    def clean(self):
        if self.searchkit_model_form.is_valid():
            self.instance.contenttype = self.searchkit_model_form.cleaned_data['searchkit_model']
        if self.formset.is_valid():
            self.instance.data = self.formset.cleaned_data
        return super().clean()


class SearchkitModelForm(forms.Form):
    """
    Form to select a content type.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        models = [m for m in apps.get_models() if is_searchable_model(m)]
        ids = [ContentType.objects.get_for_model(m).id for m in models]
        queryset = self.fields['searchkit_model'].queryset.filter(pk__in=ids)
        self.fields['searchkit_model'].queryset = queryset

    searchkit_model = forms.ModelChoiceField(
        queryset=ContentType.objects.all().order_by('app_label', 'model'),
        label=_('Model'),
        empty_label=_('Select a Model'),
        widget=forms.Select(attrs={
            "class": RELOAD_CSS_CLASS,
            "data-reload-handler": "change",
            "data-total-forms": 1,
        }),
    )


class FieldPlan:
    """
    A class holding all the logic to build a dynamic searchkit form. This
    includes building the choices for the lookup and operator form fields as
    well as choosing the right form field for the value.

    There are also some meta data included to build the media property of the
    searchkit formset.
    """

    OPERATOR_DESCRIPTION = {
        'isnull': _('is null'),
        'exact': _('is exact'),
        'contains': _('contains'),
        'startswith': _('starts with'),
        'endswith': _('ends with'),
        'regex': _('matches regular expression'),
        'gt': _('is greater than'),
        'gte': _('is greater than or equal'),
        'lt': _('is lower than'),
        'lte': _('is lower than or equal'),
        'range': _('is within range'),
        'in': _('is one of'),
    }

    TRUE_FALSE_CHOICES = (
        (True, _('Yes')),
        (False, _('No'))
    )

    CHARACTER_FIELD_TYPES = (
        models.CharField,
        models.TextField,
        models.EmailField,
        models.URLField,
        models.UUIDField,  # TODO: Does this work with standard operators?
    )

    ARITHMETIC_FIELD_TYPES = (
        models.IntegerField,
        models.BigIntegerField,
        models.DecimalField,
        models.FloatField,
        models.DateField,
        models.TimeField,
        models.DateTimeField,
    )

    SUPPORTED_FIELD_TYPES = (
        models.BooleanField,
        *CHARACTER_FIELD_TYPES,
        *ARITHMETIC_FIELD_TYPES,
    )

    RANGE_FORM_FIELD_TYPES = (
        skfields.IntegerRangeField,
        skfields.DecimalRangeField,
        skfields.FloatRangeField,
        skfields.DateRangeField,
        skfields.TimeRangeField,
        skfields.DateTimeRangeField,
    )


    def __init__(self, model, initial=None):
        self.model = model
        self.model_tree = ModelTree(model)
        self.initial = initial or dict()
        self.field_lookup = None
        self.model_field = None

    def _get_model_field(self):
        # Split the lookup. The last piece will be the field name. Everything
        # else are relational fields.
        path = self.field_lookup.split('__')

        # If we have a path to a related model we use the model tree.
        if len(path) > 1:
            model = self.model_tree.get('__'.join(path[:-1])).model
        else:
            model = self.model

        return model._meta.get_field(path[-1])

    def get_field_lookup_choices(self):
        # Not all fields have a verbose_name attribute.
        get_field_name = lambda f: getattr(f, 'verbose_name', f.name)
        relations = dict(
            one_to_one='one-to-one',
            many_to_one='many-to-one',
            one_to_many='one-to-many',
            many_to_many='many-to-many',
        )
        choices = []

        # Iterate the model tree...
        for node in self.model_tree.iterate():

            # Create a new option group for each model.
            if node.is_root:
                opt_group = (None, [])
            else:
                relation = [s for r, s in relations.items() if getattr(node.field, r)][0]
                group_label = ' -> '.join([f'`{get_field_name(n.field)}`' for n in node.path[1:]])
                group_label += f' => {node.model._meta.app_label.title()} | {node.model._meta.verbose_name} ({relation})'
                opt_group = (group_label, [])

            # Loop the model fields to build the option group.
            for model_field in node.model._meta.fields:

                # Skip unsupported fields.
                if not isinstance(model_field, self.SUPPORTED_FIELD_TYPES):
                    continue

                if node.is_root:
                    lookup = model_field.name
                else:
                    lookup = f'{node.field_path}__{model_field.name}'

                opt_group[1].append((lookup, model_field.verbose_name))

            # Append the option group to the choices.
            choices.append(opt_group)

        return choices

    def get_operator_choices(self, field_lookup):
        self.field_lookup = field_lookup
        self.model_field = self._get_model_field()

        if isinstance(self.model_field, models.BooleanField):
            operators = ['exact']

        elif isinstance(self.model_field, models.TextField):
            operators = ['contains', 'startswith', 'endswith', 'regex']

        elif isinstance(self.model_field, self.CHARACTER_FIELD_TYPES):
            operators = ['exact', 'contains', 'startswith', 'endswith', 'regex', 'in']

        elif isinstance(self.model_field, self.ARITHMETIC_FIELD_TYPES):
            operators = ['exact', 'gt', 'gte', 'lt', 'lte', 'range']

        # Add an isnull lookup for model fields allowing null values. Exclude
        # boolean fields since they are handled with a null boolean form field.
        if self.model_field.null and not isinstance(self.model_field, models.BooleanField):
            operators = [*operators, 'isnull']

        return [(None, [(o, self.OPERATOR_DESCRIPTION[o]) for o in operators])]

    def get_form_field(self, operator):
        model_field_class = type(self.model_field)

        # Use a simple boolean form field for the isnull operator.
        if operator == 'isnull':
            form_field = forms.BooleanField(
                required=False,
                widget=forms.Select(choices=self.TRUE_FALSE_CHOICES),
                )

        # Create form field for character based field types.
        elif isinstance(self.model_field, self.CHARACTER_FIELD_TYPES):

            # With these operators we use a standard search term field.
            if operator in ['contains', 'startswith', 'endswith', 'regex']:
                form_field = forms.CharField(widget=widgets.AdminTextInputWidget)

            # Use a choice field for exact and in operators.
            # FIXME: We might should use a simple ChoiceField if the model's
            # field have choices defined.
            elif operator in ['exact', 'in']:
                if operator == 'exact':
                    form_field = skfields.Select2Field(self.model_field)
                elif operator == 'in':
                    form_field = skfields.MultiSelect2Field(self.model_field)

        # Handle arithmentic based field types.
        elif isinstance(self.model_field, self.ARITHMETIC_FIELD_TYPES):
            # Use range fields for the range operator.
            if operator == 'range':

                # We choose the appropriate range field based on the class names.
                test = lambda k: k.__name__.replace('Range', '') == model_field_class.__name__
                try:
                    klass = [k for k in self.RANGE_FORM_FIELD_TYPES if test(k)][0]

                # Some model fields like AutoField have no equivalent range
                # field. Those are fine with the IntegerRangeField.
                except IndexError:
                    klass = skfields.IntegerRangeField

                form_field = klass()

            # For exact operator only use a choice field if the model field has
            # choices.
            elif operator == 'exact' and self.model_field.choices:
                form_field = forms.ChoiceField(choices=self.model_field.choices)

            # TODO: Check the core code on how they use the defaults.

            # For all other operators we use the field specific default.
            else:
                # These django-admin default definition specifies widgets and
                # form classes which are utilized by the django admin site. This
                # way we get access to the famous calender widget of
                # django-admin for date and datetime fields.
                defaults = FORMFIELD_FOR_DBFIELD_DEFAULTS.get(model_field_class, dict()).copy()

                # Do we have a form class within the defaults? (True for the
                # split-date-time-field.)
                klass = defaults.pop('form_class', None)

                # Otherwise get the type of the formfield returned by the model
                # field.
                _form_field = self.model_field.formfield()
                klass = klass or type(_form_field) if _form_field else None

                # Model fields as AutoField and BigAutoField return None for
                # formfield(). So we use an IntegerField as backup.
                klass = klass or forms.IntegerField

                # Initialize the formfield.
                form_field = klass(**defaults)

        # Handle BooleanField
        elif isinstance(self.model_field, models.BooleanField):
            # The formfield method is aware of the null attribute and returns a
            # null-boolean form field.
            if self.model_field.null:
                form_field = self.model_field.formfield()

            # Otherwise we use a Select widget with True and False.
            else:
                form_field = forms.BooleanField(
                    required=False,
                    widget=forms.Select(choices=self.TRUE_FALSE_CHOICES),
                    )

        return form_field


class BaseSearchkitForm(forms.Form):
    """
    Searchkit form representing a queryset filter rule.

    Based on the model three fields are dynamically created:
    * The "Model field" field with model field lookup paths as choices.
    * The operator field with lookup types as choices.
    * A value field which is created based on the model field and the operator.

    Additionally there is an exclude boolean field which marks filter rules that
    should be used to exclude objects from the search result.
    """
    exclude = forms.BooleanField(required=False, label=_('Exclude from the result'))

    model = None  # Set by the formset factory.
    html_attrs = {
        "class": RELOAD_CSS_CLASS,
        "data-reload-handler": "change",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.field_plan = FieldPlan(self.model, self.initial)
        self._add_field_lookup_field()
        self._add_operator_field()
        self._add_value_field()

    @cached_property
    def unprefixed_data(self):
        data = dict()
        for key, value in self.data.items():
            if key.startswith(self.prefix):
                data[key[len(self.prefix) + 1:]] = value
        return data

    def _preload_clean_data(self, field_name):
        # Try the initial value first since it is already cleaned.
        if self.initial and field_name in self.initial:
            return self.initial[field_name]
        # Otherwise look up the data dict.
        elif field_name in self.unprefixed_data:
            try:
                # Do we have a valid value?
                return self.fields[field_name].clean(self.unprefixed_data[field_name])
            except forms.ValidationError:
                # Otherwise return the first option.
                return self.fields[field_name].choices[0][1][0][0]
        else:
            # At a default return the first option which will be the selected
            # one.
            return  self.fields[field_name].choices[0][1][0][0]

    def _add_field_lookup_field(self):
        choices = self.field_plan.get_field_lookup_choices()
        field = forms.ChoiceField(label=_('Model field'), choices=choices)
        field.widget.attrs.update(self.html_attrs)
        self.fields['field'] = field

    def _add_operator_field(self):
        field_lookup = self._preload_clean_data('field')
        choices = self.field_plan.get_operator_choices(field_lookup)
        field = forms.ChoiceField(label=_('Operator'), choices=choices)
        field.widget.attrs.update(self.html_attrs)
        self.fields['operator'] = field

    def _add_value_field(self):
        operator = self._preload_clean_data('operator')
        form_field = self.field_plan.get_form_field(operator)
        self.fields['value'] = form_field


class BaseSearchkitFormSet(forms.BaseFormSet):
    """
    Formset holding all searchkit forms.
    """
    template_name = "searchkit/searchkit.html"
    template_name_div = "searchkit/searchkit.html"
    model = None  # Set by the formset factory.

    def add_prefix(self, index):
        return "%s-%s-%s-%s" % (self.prefix, self.model._meta.app_label, self.model._meta.model_name, index)

    @classmethod
    def get_default_prefix(self):
        return "searchkit"

    @cached_property
    def forms(self):
        if self.model:
            return super().forms
        else:
            return []

    @property
    def media(self):
        # We build a media collection including everything that might be needed
        # by reloaded versions of the formset. So we do not have to dynamically
        # update media assets on the client site.

        # Basic searchkit media.
        media = forms.Media(js=[
            "searchkit/js/searchkit.js",
            "searchkit/js/widgets/datetime.js",
            "searchkit/js/widgets/select2.js",
        ])

        # Get media assets for calender and select2 widgets.
        media += widgets.AdminSplitDateTime().media
        media += widgets.AutocompleteSelect(None, None).media

        return media

    def get_context(self):
        context = super().get_context()
        context.update(
            sk_reload_css_class=RELOAD_CSS_CLASS,
            sk_reload_url=reverse('searchkit-reload'),
            sk_total_form_count=self.total_form_count,
        )
        return context


def searchkit_formset_factory(model, **kwargs):
    form = type('SearchkitForm', (BaseSearchkitForm,), dict(model=model))
    formset = type('SearchkitFormSet', (BaseSearchkitFormSet,), dict(model=model))
    return forms.formset_factory(
        form=form,
        formset=formset,
        **kwargs
    )



from collections import OrderedDict
from picklefield.fields import PickledObjectField
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext_lazy as _


class Search(models.Model):
    name = models.CharField(_('Search name'), max_length=255)
    contenttype = models.ForeignKey(ContentType, on_delete=models.CASCADE, verbose_name=_('Model'))
    data = PickledObjectField(_('Serialized filter rule data'))
    created_date = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = ('name', 'contenttype')

    def as_lookups(self):
        includes = OrderedDict()
        excludes = OrderedDict()
        for data in self.data:
            if data['exclude']:
                excludes[f'{data["field"]}__{data["operator"]}'] = data['value']
            else:
                includes[f'{data["field"]}__{data["operator"]}'] = data['value']
        return includes, excludes



from rest_framework.views import APIView
from rest_framework.exceptions import APIException
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.renderers import JSONRenderer
from rest_framework.renderers import StaticHTMLRenderer
from django.apps import apps
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import FieldError
from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import PermissionDenied
from .forms import SearchkitModelForm
from .forms import searchkit_formset_factory


class InvalidSearchkitModel(APIException):
    status_code = 400
    default_detail = _('Invalid searchkit model.')
    default_code = 'invalid_searchkit_model'


class SearchkitPermission(BasePermission):
    def has_permission(self, request, view):
        # Allow access only if the user has the 'add' or 'change' permissions.
        return (
            request.user.has_perm('searchkit.add_search')
            or request.user.has_perm('searchkit.change_search')
        )


class SearchkitView(APIView):
    """
    Update the searchkit formset via ajax.
    """
    permission_classes = [SearchkitPermission]
    renderer_classes = [StaticHTMLRenderer]

    def get(self, request, **kwargs):

        model_form = SearchkitModelForm(data=request.GET)
        if model_form.is_valid():
            model = model_form.cleaned_data['searchkit_model'].model_class()
        else:
            raise InvalidSearchkitModel(model_form.errors)

        formset = searchkit_formset_factory(model=model)(data=request.GET)
        return Response(formset.render())


class AutocompleteView(APIView):
    """
    Autocomplete view for select2 value fields.
    """
    permission_classes = [SearchkitPermission]
    renderer_classes = [JSONRenderer]
    paginate_by = 25

    def get(self, request, **kwargs):
        try:
            app_label = request.GET['sk_autocomplete_app_label']
            model_name = request.GET['sk_autocomplete_model_name']
            field_name = request.GET['sk_autocomplete_field_name']
        except KeyError as e:
            raise PermissionDenied from e

        try:
            model = apps.get_model(app_label, model_name)
        except LookupError as e:
            raise PermissionDenied from e

        try:
            field = model._meta.get_field(field_name)
        except FieldDoesNotExist as e:
            raise PermissionDenied from e

        perm = f'{app_label}.view_{model_name}'
        if not request.user.has_perm(perm):
            msg = f"User {request.user} is not allowed to view {model_name}"
            raise PermissionDenied(msg)

        # We use distinct together with an order by the field in question. This
        # neutralizes all former order_by clauses which might add columns to the
        # sql distinct statement. See the django docs for more details.
        queryset = model.objects.values_list(field.attname, flat=True)
        queryset = queryset.order_by(field.attname).distinct()

        if term := request.GET.get('term'):
            # FIXME: How to handle very big search results?
            queryset = queryset.filter(**{f'{field.attname}__icontains': term})
            more = False

        else:
            page = int(request.GET.get('page', 1))
            start = (page - 1) * self.paginate_by
            end = page * self.paginate_by
            count = queryset.count()
            queryset = queryset[start:end]
            more = count > end

        result = dict(
            results=[dict(id=v, text=v) for v in queryset],
            pagination=dict(more=more),
        )
        return Response(result)



from django.db import models


CHARS_CHOICES = (
    ('one', 'One'),
    ('two', 'Two'),
    ('three', 'Three'),
    ('four', 'Four'),
)
INTEGER_CHOICES = (
    (1, 1),
    (2, 2),
    (3, 3),
    (4, 4),
)


class ModelA(models.Model):
    boolean = models.BooleanField(null=True)
    chars = models.CharField(max_length=255)
    chars_choices = models.CharField(max_length=255, choices=CHARS_CHOICES)
    text = models.TextField()
    email = models.EmailField()
    url = models.URLField()
    uuid = models.UUIDField()
    integer = models.IntegerField()
    big_integer = models.BigIntegerField()
    integer_choices = models.IntegerField(choices=INTEGER_CHOICES)
    float = models.FloatField()
    decimal = models.DecimalField(max_digits=5, decimal_places=2)
    date = models.DateField()
    time = models.TimeField()
    datetime = models.DateTimeField()
    model_b = models.OneToOneField('ModelB', on_delete=models.CASCADE, null=True, blank=True)
    model_d = models.ManyToManyField('ModelD')


class ModelB(models.Model):
    chars = models.CharField(max_length=255, null=True, blank=True)
    integer = models.IntegerField(null=True, blank=True)
    decimal = models.DecimalField(max_digits=5, decimal_places=2, null=True, blank=True)
    date = models.DateField(null=True, blank=True)
    datetime = models.DateTimeField(null=True, blank=True)
    model_c = models.ForeignKey('ModelC', on_delete=models.CASCADE)


class ModelC(models.Model):
    boolean = models.BooleanField(null=False, default=True)
    chars = models.CharField(max_length=255)
    integer = models.IntegerField()
    date = models.DateField()


class ModelD(models.Model):
    chars = models.CharField(max_length=255)
    integer = models.IntegerField()
    date = models.DateField()



from django import forms
from django.urls import reverse
from django.utils.http import urlencode
from django.contrib.admin import widgets
from django.forms.models import ModelChoiceIterator
from django.utils.translation import gettext_lazy as _
from . import widgets as skwidgets


class BaseRangeField(forms.MultiValueField):
    field_type = None
    widget_type = None
    error_message = _("Enter 'From' and 'To' value of the range.")
    default_error_messages = {
        "invalid": error_message,
        "incomplete": error_message,
    }

    def __init__(self, **kwargs):
        widget = skwidgets.RangeWidget(widgets=[self.widget_type(), self.widget_type()])
        fields = (self.field_type(), self.field_type())
        super().__init__(
            fields=fields,
            widget=widget,
            require_all_fields=True,
            **kwargs
        )

    def compress(self, data_list):
        """We want the data list as data list."""
        return data_list


class IntegerRangeField(BaseRangeField):
    field_type = forms.IntegerField
    widget_type = forms.NumberInput


class FloatRangeField(IntegerRangeField):
    field_type = forms.FloatField


class DecimalRangeField(IntegerRangeField):
    field_type = forms.DecimalField


class DateRangeField(BaseRangeField):
    field_type = forms.DateField
    widget_type = widgets.AdminDateWidget


class TimeRangeField(BaseRangeField):
    field_type = forms.TimeField
    widget_type = widgets.AdminTimeWidget


class DateTimeRangeField(BaseRangeField):
    field_type = forms.SplitDateTimeField
    widget_type = widgets.AdminSplitDateTime


class FieldChoiceIterator(ModelChoiceIterator):
    """
    The ModelChoiceIterator do what we need. We just build our choices based
    on a field value and not an object.
    """
    def choice(self, value):
        return (value, value)


class Select2Mixin:
    """
    Get choices based on the distinct values of a model field.
    We took some code and inspiration from django's ModelChoiceField.
    """
    iterator = FieldChoiceIterator

    def __init__(self, model_field, *args, **kwargs):
        self.model_field = model_field
        self.model = model_field.model
        self.queryset = self._get_queryset()
        # ModelChoiceIterator expects an empty_label attribute.
        self.empty_label = None
        super().__init__(*args, **kwargs)

    def _get_queryset(self):
        lookup = self.model_field.attname
        queryset = self.model.objects.all()
        queryset = queryset.values_list(lookup, flat=True)
        # We order by our field to neutralize former ordering which might
        # interfere with the sql distinct statement.
        queryset = queryset.order_by(lookup)
        queryset = queryset.distinct()
        return queryset

    def _set_queryset(self, queryset):
        self._queryset = None if queryset is None else queryset.all()
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)

    def _get_choices(self):
        # FIXME: This comes from django core. But I wonder if the choices are
        # accessed multiple times, won't we evaluate the queryset each time a
        # new? It does make sense to me yet.


        # Otherwise, execute the QuerySet in self.queryset to determine the
        # choices dynamically. Return a fresh ModelChoiceIterator that has not been
        # consumed. Note that we're instantiating a new ModelChoiceIterator *each*
        # time _get_choices() is called (and, thus, each time self.choices is
        # accessed) so that we can ensure the QuerySet has not been consumed. This
        # construct might look complicated but it allows for lazy evaluation of
        # the queryset.
        return self.iterator(self)

    choices = property(_get_choices, forms.ChoiceField.choices.fset)

    def _get_url(self):
        base_url = reverse('searchkit-autocomplete')
        url_params = {
            'sk_autocomplete_app_label': self.model._meta.app_label,
            'sk_autocomplete_model_name': self.model._meta.model_name,
            'sk_autocomplete_field_name': self.model_field.attname,
        }
        return f'{base_url}?{urlencode(url_params)}'

    def widget_attrs(self, widget):
        """
        Set select2's AJAX attributes.

        Attributes can be set using the html5 data attribute.
        Nested attributes require a double dash as per
        https://select2.org/configuration/data-attributes#nested-subkey-options
        """
        attrs = super().widget_attrs(widget)
        attrs.update({
            "class": 'admin-autocomplete',
            "data-placeholder": "search...",
            "data-theme": "admin-autocomplete",
            "data-ajax--dataType": 'json',
            "data-allow-clear": 'false',
            "data-ajax--type": "GET",
            "data-ajax--delay": 250,
            "data-ajax--url": self._get_url(),
            "data-ajax--cache": "true",
        })
        return attrs


class Select2Field(Select2Mixin, forms.ChoiceField):
    widget = skwidgets.Select2


class MultiSelect2Field(Select2Mixin, forms.MultipleChoiceField):
    widget = skwidgets.MultiSelect2



from django.template import Library
from django.contrib.admin.helpers import Fieldset


register = Library()

@register.inclusion_tag("admin/includes/fieldset.html")
def as_fieldset(form, heading_level=2, prefix='', id_prefix=0, id_suffix='', **fieldset_kwargs):
    """
    Create and render a fieldset for form.
    """
    fieldset = Fieldset(form, fields=form.fields, **fieldset_kwargs)
    context = dict(
        fieldset=fieldset,
        heading_level=heading_level,
        prefix=prefix,
        id_prefix=id_prefix,
        id_suffix=id_suffix,
    )
    return context



import os, sys, json
from pprint import pprint
from decimal import Decimal
from contextlib import contextmanager
from urllib.parse import urlencode
from django.test import TestCase
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
from django.urls import reverse
from example.models import ModelA
from example.management.commands.createtestdata import Command as CreateTestData
from searchkit.forms import FieldPlan
from searchkit.utils import ModelTree
from searchkit.forms import SearchForm
from searchkit.forms import SearchkitModelForm
from searchkit.forms import BaseSearchkitFormSet
from searchkit.forms import searchkit_formset_factory
from searchkit.models import Search
from searchkit.views import AutocompleteView
from searchkit import __version__


# Check django version.
import django
print(f'DJANGO-VERSION: {django.VERSION}')


SearchkitFormSet = searchkit_formset_factory(model=ModelA)
SearchkitForm = SearchkitFormSet.form


INITIAL_DATA = [
    dict(
        field='id',
        operator='gt',
        value=123,
    ),
    dict(
        field='id',
        operator='range',
        value=[123, 432],
    ),
    dict(
        field='boolean',
        operator='exact',
        value=True,
    ),
    dict(
        field='chars',
        operator='in',
        value=['ModelA chars 1', 'ModelA chars 2'],
    ),
    dict(
        field='chars_choices',
        operator='exact',
        value='one',
    ),
    dict(
        field='text',
        operator='contains',
        value='xyz',
    ),
    dict(
        field='email',
        operator='startswith',
        value='user12',
    ),
    dict(
        field='url',
        operator='regex',
        value='^.+\.com/[a-z]+/66$',
    ),
    dict(
        field='uuid',
        operator='endswith',
        value='x',
    ),
    dict(
        field='integer',
        operator='range',
        value=[1, 123],
    ),
    dict(
        field='big_integer',
        operator='gt',
        value=1000,
    ),
    dict(
        field='integer_choices',
        operator='exact',
        value=3,
    ),
    dict(
        exclude=True,
        field='float',
        operator='lte',
        value=10.02,
    ),
    dict(
        field='decimal',
        operator='gte',
        value=Decimal('2.2'),
    ),
    dict(
        field='date',
        operator='gt',
        value='2020-05-14',
    ),
    dict(
        field='time',
        operator='lte',
        value='22:45',
    ),
    dict(
        field='datetime',
        operator='exact',
        value=['2025-05-14', '08:45'],
    ),
    dict(
        field='model_b__chars',
        operator='isnull',
        value=False,
    ),
    dict(
        field='model_b__model_c__integer',
        operator='gt',
        value=4,
    ),
]

add_prefix = lambda i: SearchkitFormSet().add_prefix(i)
MODELA_CT = ContentType.objects.get_for_model(ModelA)
DEFAULT_PREFIX = SearchkitFormSet.get_default_prefix()

def get_form_data(initial_data=INITIAL_DATA):
    count = len(initial_data)
    data = {
        'name': 'test search',                          # The name of the search.
        'searchkit_model': f'{MODELA_CT.pk}',         # Data for the searchkit-model form.
        f'{DEFAULT_PREFIX}-TOTAL_FORMS': f'{count}',    # Data for the managment form.
        f'{DEFAULT_PREFIX}-INITIAL_FORMS': f'{count}',  # Data for the managment form.
    }
    for i, idata in enumerate(initial_data):
        prefix = SearchkitFormSet().add_prefix(i)
        for key, value in idata.items():
            # Create multiple value fields for range operators or the datetime
            # field. Exclude the in operator since it expects a list as value.
            if isinstance(value, list) and idata['operator'] != 'in':
                data.update({f'{prefix}-{key}_{0}': value[0]})
                data.update({f'{prefix}-{key}_{1}': value[1]})
            else:
                data.update({f'{prefix}-{key}': value})
    return data

FORM_DATA = get_form_data()


@contextmanager
def silence_stdout():
    old_stdout = sys.stdout
    try:
        with open(os.devnull, "w") as dev_null:
            sys.stdout = dev_null
            yield dev_null
    finally:
        sys.stdout = old_stdout



class CheckFormMixin:
    """
    Mixin to check the form fields and their choices.
    """
    def check_form(self, form):
        # Three fields should be generated on instantiation.
        self.assertIn('field', form.fields)
        self.assertIn('operator', form.fields)
        self.assertIn('value', form.fields)
        self.assertEqual(len(form.fields), 4)

        # Check field choices for the model.
        form_model_field = form.fields['field']
        self.assertTrue(form_model_field.choices)
        options = [c[0] for g in form_model_field.choices for c in g[1]]
        tree = ModelTree(ModelA)
        for node in tree.iterate():
            for model_field in node.model._meta.fields:
                if not any(isinstance(model_field, f) for f in FieldPlan.SUPPORTED_FIELD_TYPES):
                    continue
                if node.is_root:
                    self.assertIn(model_field.name, options)
                else:
                    self.assertIn(f'{node.field_path}__{model_field.name}', options)

        # Check operator field has choices.
        self.assertTrue(form.fields['operator'].choices)


class CreateTestDataMixin:
    @classmethod
    def setUpTestData(cls):
        with silence_stdout():
            CreateTestData().handle()


class SearchkitFormTestCase(CheckFormMixin, TestCase):

    def test_blank_searchkitform(self):
        form = SearchkitForm(prefix=add_prefix(0))
        self.check_form(form)

        # Form should not be bound or valid.
        self.assertFalse(form.is_bound)
        self.assertFalse(form.is_valid())

    def test_searchkitform_with_invalid_model_field_data(self):
        data = {
            f'{add_prefix(0)}-field': 'foobar',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be invalid.
        self.assertFalse(form.is_valid())

        # Check error message in html.
        errors = ['Select a valid choice. foobar is not one of the available choices.']
        self.assertIn(errors, form.errors.values())

    def test_searchkitform_with_valid_model_field_data(self):
        data = {
            f'{add_prefix(0)}-field': 'integer',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be invalid since no value data is provieded.
        self.assertFalse(form.is_valid())

    def test_searchkitform_with_invalid_operator_data(self):
        data = {
            f'{add_prefix(0)}-field': 'integer',
            f'{add_prefix(0)}-operator': 'foobar',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be invalid.
        self.assertFalse(form.is_valid())

        # Check error message in html.
        errors = ['Select a valid choice. foobar is not one of the available choices.']
        self.assertIn(errors, form.errors.values())

    def test_searchkitform_with_valid_operator_data(self):
        data = {
            f'{add_prefix(0)}-field': 'integer',
            f'{add_prefix(0)}-operator': 'exact',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be invalid since no value data is provieded.
        self.assertFalse(form.is_valid())

    def test_searchkitform_with_valid_data(self):
        data = {
            f'{add_prefix(0)}-field': 'integer',
            f'{add_prefix(0)}-operator': 'exact',
            f'{add_prefix(0)}-value': '123',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be valid.
        self.assertTrue(form.is_valid())

    def test_searchkitform_with_invalid_data(self):
        data = {
            f'{add_prefix(0)}-field': 'integer',
            f'{add_prefix(0)}-operator': 'exact',
            f'{add_prefix(0)}-value': 'foobar',
        }
        form = SearchkitForm(data, prefix=add_prefix(0))
        self.check_form(form)

        # Form should be invalid.
        self.assertFalse(form.is_valid())

        # Check error message in html.
        errors = ['Enter a whole number.']
        self.assertIn(errors, form.errors.values())


class SearchkitFormSetTestCase(CreateTestDataMixin, CheckFormMixin, TestCase):
    def test_blank_searchkitform(self):
        # Instantiating the formset neither with a model instance nor with model
        # related data or initial data should result in a formset without forms,
        # that is invalid and unbound.
        formset = SearchkitFormSet()
        self.assertFalse(formset.is_bound)
        self.assertFalse(formset.is_valid())

    def test_searchkit_formset_with_valid_data(self):
        formset = SearchkitFormSet(FORM_DATA)
        self.assertTrue(formset.is_valid())

    def test_searchkit_formset_with_invalid_data(self):
        data = FORM_DATA.copy()
        del data[f'{add_prefix(0)}-value']
        formset = SearchkitFormSet(data)
        self.assertFalse(formset.is_valid())

        # Check error message in html.
        errors = ['This field is required.']
        self.assertIn(errors, formset.forms[0].errors.values())

    def test_searchkit_formset_with_initial_data(self):
        formset_class = searchkit_formset_factory(model=ModelA, extra=0)
        formset = formset_class(initial=INITIAL_DATA)
        self.assertFalse(formset.is_bound)
        self.assertFalse(formset.is_valid())
        self.assertEqual(len(formset.forms), len(INITIAL_DATA))
        for i, form in enumerate(formset.forms):
            self.assertEqual(form.initial, INITIAL_DATA[i])
            self.check_form(form)


class SearchkitSearchFormTestCase(CreateTestDataMixin, TestCase):
    def test_searchkit_search_form_without_data(self):
        form = SearchForm()
        self.assertFalse(form.is_bound)
        self.assertFalse(form.is_valid())
        self.assertIsInstance(form.formset, BaseSearchkitFormSet)
        self.assertEqual(form.formset.model, None)

    def test_searchkit_search_form_with_data(self):
        form = SearchForm(FORM_DATA)
        self.assertTrue(form.is_bound)
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.formset, BaseSearchkitFormSet)
        self.assertEqual(form.formset.model, ModelA)
        self.assertEqual(form.instance.data, form.formset.cleaned_data)

        # Saving the instance works.
        form.instance.save()
        self.assertTrue(form.instance.pk)

        # Using the instance data as filter rules works.
        includes, excludes = form.instance.as_lookups()
        self.assertEqual(len(includes) + len(excludes), len(INITIAL_DATA))
        for data in INITIAL_DATA:
            if data.get('exclude'):
                self.assertIn(f"{data['field']}__{data['operator']}", excludes)
            else:
                self.assertIn(f"{data['field']}__{data['operator']}", includes)


class SearchkitModelFormTestCase(TestCase):
    def test_searchkit_model_form_choices(self):
        form = SearchkitModelForm()
        labels = [c[1] for c in form.fields['searchkit_model'].choices]
        self.assertEqual(len(labels), 4)
        self.assertEqual('select a model', labels[0].lower())
        self.assertEqual('example | model a', labels[1].lower())
        self.assertEqual('example | model b', labels[2].lower())
        self.assertEqual('example | model c', labels[3].lower())


class AdminBackendTest(CreateTestDataMixin, TestCase):

    def setUp(self):
        admin = User.objects.get(username='admin')
        self.client.force_login(admin)

    def test_search_form(self):
        url = reverse('admin:searchkit_search_add')
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        select = '<select name="searchkit_model" class="searchkit-reload" data-reload-handler="change" data-total-forms="1" required id="id_searchkit_model">'
        for snippet in select.split(' '):
            self.assertIn(snippet, str(resp.content))

    def test_search_form_with_initial(self):
        url = reverse('admin:searchkit_search_add') + f'?searchkit_model={MODELA_CT.id}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        select = '<select name="searchkit_model" class="searchkit-reload" data-reload-handler="change" data-total-forms="1" required id="id_searchkit_model">'
        for snippet in select.split(' '):
            self.assertIn(snippet, str(resp.content))
        self.assertIn(f'<option value="{MODELA_CT.id}" selected>', str(resp.content))
        self.assertIn('name="searchkit-example-modela-0-field"', str(resp.content))

    def test_add_search(self):
        # Create a search object via the admin backend.
        url = reverse('admin:searchkit_search_add')
        data = FORM_DATA.copy()
        data['_save_and_apply'] = True
        resp = self.client.post(url, data, follow=True)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(Search.objects.all()), 1)

        # Load the saved search.
        url = reverse('admin:searchkit_search_change', args=(1,))
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertIn('<select name="searchkit_model"', str(resp.content))
        self.assertIn(f'<option value="{MODELA_CT.id}" selected>', str(resp.content))
        self.assertIn('name="searchkit-example-modela-0-field"', str(resp.content))
        self.assertIn('value="ModelA chars 1"', str(resp.content))

        # Change it via admin backend.
        data['name'] = 'Changed name'
        data['searchkit-example-modela-0-field'] = 'boolean'
        data['searchkit-example-modela-0-operator'] = 'exact'
        data['searchkit-example-modela-0-value'] = 'true'
        resp = self.client.post(url, data, follow=True)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(Search.objects.get(pk=1).name, data['name'])

        # Will the search be listed in the admin filter?
        url = reverse('admin:example_modela_changelist')
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertIn('href="?search=1"', str(resp.content))
        self.assertIn(data['name'], str(resp.content))


class SearchkitViewTest(CreateTestDataMixin, TestCase):

    def setUp(self):
        admin = User.objects.get(username='admin')
        self.client.force_login(admin)

        self.initial = [
            dict(
                field='integer',
                operator='exact',
                value=1,
            )
        ]
        self.initial_range = [
            dict(
                field='integer',
                operator='range',
                value=[1,3],
            )
        ]

    def test_searchkit_view_with_anonymous_user(self):
        self.client.logout()
        data = get_form_data(self.initial)
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 403)

    def test_searchkit_view_invalid_data(self):
        initial = self.initial.copy()
        initial[0]['value'] = 'no integer'
        data = get_form_data(initial)
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        html_error = '<li>Enter a whole number.</li>'
        self.assertInHTML(html_error, str(resp.content))

    def test_searchkit_view_missing_data(self):
        initial = self.initial.copy()
        del(initial[0]['value'])
        data = get_form_data(initial)
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        html_error = '<li>This field is required.</li>'
        self.assertInHTML(html_error, str(resp.content))

    def test_searchkit_view_with_range_operator(self):
        data = get_form_data(self.initial_range)
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        html = '<input type="number" name="searchkit-example-modela-0-value_1" value="3" id="id_searchkit-example-modela-0-value_1">'
        self.assertInHTML(html, resp.content.decode('utf-8'))

    def test_searchkit_view_with_model(self):
        data = get_form_data(self.initial)
        data['searchkit_model'] = ContentType.objects.get_for_model(ModelA).pk
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)

    def test_searchkit_view_with_invalid_model(self):
        data = get_form_data(self.initial)
        data['searchkit_model'] = 9999  # Non-existing content type.
        url_params = urlencode(data)
        base_url = reverse('searchkit-reload')
        url = f'{base_url}?{url_params}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 400)


class Select2ViewTestCase(CreateTestDataMixin, TestCase):
    def setUp(self):
        admin = User.objects.get(username='admin')
        self.client.force_login(admin)
        self.url = reverse('searchkit-autocomplete')
        self.data = {
            'sk_autocomplete_app_label': ModelA._meta.app_label,
            'sk_autocomplete_model_name': ModelA._meta.model_name,
            'sk_autocomplete_field_name': 'chars',
        }

    def test_select2_view_with_anonymous_user(self):
        self.client.logout()
        data = self.data.copy()
        url = f'{self.url}?{urlencode(self.data)}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 403)

    def test_select2_view_without_data(self):
        resp = self.client.get(self.url)
        self.assertEqual(resp.status_code, 403)

    def test_select2_view_with_base_data(self):
        url = f'{self.url}?{urlencode(self.data)}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        result = json.loads(resp.content)
        self.assertTrue('pagination' in result)
        self.assertTrue(result['pagination']['more'])
        self.assertTrue('results' in result)
        self.assertEqual(len(result['results']), AutocompleteView.paginate_by)

    def test_select2_view_with_paging(self):
        data = self.data.copy()
        data['page'] = 10
        url = f'{self.url}?{urlencode(data)}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        result = json.loads(resp.content)
        self.assertTrue('pagination' in result)
        self.assertTrue(result['pagination']['more'])
        self.assertTrue('results' in result)
        self.assertEqual(len(result['results']), AutocompleteView.paginate_by)

    def test_select2_view_with_search_term(self):
        data = self.data.copy()
        data['term'] = 'Model'
        url = f'{self.url}?{urlencode(data)}'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        result = json.loads(resp.content)
        self.assertTrue('pagination' in result)
        self.assertFalse(result['pagination']['more'])
        self.assertTrue('results' in result)
        field_name = self.data['sk_autocomplete_field_name']
        lookup = f'{field_name}__icontains'
        queryset = ModelA.objects.order_by(lookup)
        queryset = queryset.values_list(field_name)
        queryset = queryset.filter(**{lookup:data['term']})
        queryset = queryset.distinct()
        self.assertEqual(len(result['results']), queryset.count())




from django import forms


class RangeWidget(forms.MultiWidget):
    """
    Range widget based on the MultiWidget with two sub-widgets.
    """
    template_name = "django/forms/widgets/rangewidget.html"
    def decompress(self, value):
        # For an empty value return a list with two None values.
        if value:
            return value
        else:
            return [None, None]


class Select2Mixin:
    """
    Prevent the widget from rendering all choices but only initial values.
    """
    def optgroups(self, name, value, attr=None):
        """
        Return options based on the value passed in - ignoring the choices at all.
        """
        # FIXME: Should we check the value against the choices?
        options = [self.create_option(name, v, v, True, i) for i, v in enumerate(value)]
        return [(None, options, 0)]


class Select2(Select2Mixin, forms.Select):
    pass


class MultiSelect2(Select2Mixin, forms.SelectMultiple):
    pass


from django.urls import path
from .views import SearchkitView
from .views import AutocompleteView


urlpatterns = [
    path("reload/", SearchkitView.as_view(), name="searchkit-reload"),
    path("autocomplete/", AutocompleteView.as_view(), name="searchkit-autocomplete"),
]



# Generated by Django 5.2.4 on 2025-07-14 12:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('example', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ModelD',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chars', models.CharField(max_length=255)),
                ('integer', models.IntegerField()),
                ('date', models.DateField()),
            ],
        ),
        migrations.AddField(
            model_name='modela',
            name='model_d',
            field=models.ManyToManyField(to='example.modeld'),
        ),
    ]



#!/usr/bin/env python

import os
from pathlib import Path
from setuptools import setup
from setuptools import find_namespace_packages


def get_version():
    """Get the local package version."""
    namespace = {}
    path = Path("searchkit", "__version__.py")
    exec(path.read_text(), namespace)
    return namespace["__version__"]


def read(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, encoding="utf-8") as file:
        return file.read()


get_version = get_version()
if 'dev' in get_version:
    dev_status = 'Development Status :: 3 - Alpha'
elif 'beta' in get_version:
    dev_status = 'Development Status :: 4 - Beta'
else:
    dev_status = 'Development Status :: 5 - Production/Stable'


setup(
    name="django-searchkit",
    version=get_version,
    description="Finally a real searchkit for django!",
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    author="Thomas Leichtfuß",
    author_email="thomas.leichtfuss@posteo.de",
    url="https://github.com/thomst/django-searchkit",
    license="BSD License",
    platforms=["OS Independent"],
    packages=find_namespace_packages(exclude=["example"]),
    include_package_data=True,
    install_requires=[
        "Django>=4.0",
        "django-picklefield",
        "djangorestframework",
        "django-modeltree",
    ],
    classifiers=[
        dev_status,
        "Framework :: Django",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Framework :: Django :: 5.1",
        "Framework :: Django :: 5.2",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    zip_safe=True,
)



# Generated by Django 5.1.3 on 2025-07-10 10:11

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ModelB',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chars', models.CharField(max_length=255)),
                ('integer', models.IntegerField()),
                ('decimal', models.DecimalField(decimal_places=2, max_digits=5)),
                ('date', models.DateField()),
                ('datetime', models.DateTimeField()),
            ],
        ),
        migrations.CreateModel(
            name='ModelC',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chars', models.CharField(max_length=255)),
                ('integer', models.IntegerField()),
                ('date', models.DateField()),
            ],
        ),
        migrations.CreateModel(
            name='ModelA',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('boolean', models.BooleanField(null=True)),
                ('chars', models.CharField(max_length=255)),
                ('chars_choices', models.CharField(choices=[('one', 'One'), ('two', 'Two'), ('three', 'Three'), ('four', 'Four')], max_length=255)),
                ('text', models.TextField()),
                ('email', models.EmailField(max_length=254)),
                ('url', models.URLField()),
                ('uuid', models.UUIDField()),
                ('integer', models.IntegerField()),
                ('big_integer', models.BigIntegerField()),
                ('integer_choices', models.IntegerField(choices=[(1, 1), (2, 2), (3, 3), (4, 4)])),
                ('float', models.FloatField()),
                ('decimal', models.DecimalField(decimal_places=2, max_digits=5)),
                ('date', models.DateField()),
                ('time', models.TimeField()),
                ('datetime', models.DateTimeField()),
                ('model_b', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='example.modelb')),
            ],
        ),
        migrations.AddField(
            model_name='modelb',
            name='model_c',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='example.modelc'),
        ),
    ]



# Generated by Django 5.1.3 on 2025-05-22 19:04

import django.db.models.deletion
import picklefield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='SearchkitSearch',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='Search name')),
                ('data', picklefield.fields.PickledObjectField(editable=False, verbose_name='Serialized filter rule data')),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('contenttype', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Model')),
            ],
            options={
                'unique_together': {('name', 'contenttype')},
            },
        ),
    ]



# Generated by Django 5.1.3 on 2025-06-07 05:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('searchkit', '0001_initial'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='SearchkitSearch',
            new_name='Search',
        ),
    ]
